from pysc2.lib import actions, features, units
from pysc2.agents import base_agent
import MacroAction
import random as rand

class SCAction():
     # Declare Constants used us PySC21
    _PLAYER_RELATIVE = features.SCREEN_FEATURES.player_relative.index
    _PLAYER_FRIENDLY = 1
    _PLAYER_NEUTRAL = 3
    _PLAYER_HOSTILE = 4

    #Move_screen actually moves the unit you have selected
    #_MOVE_SCREEN = actions.RAW_FUNCTIONS.Move_screen.id
    #_SELECT_ARMY = actions.RAW_FUNCTIONS.select_army.id
    #_SELECT_POINT = actions.RAW_FUNCTIONS.select_point.id
    #_ATTACK_SCREEN = actions.RAW_FUNCTIONS.Attack_screen.id
    #_NO_OP = actions.RAW_FUNCTIONS.no_op.id
    #_BUILD_SUPPLY_DEPOT = actions.RAW_FUNCTIONS.Build_SupplyDepot_pt.id

    #Unit Type constants1
    _SCV = 45
    _MINERAL_FIELD = 341
    _VESPENE_GEYSER = 342
    _COMMAND_CENTER = 18
    _REFINERY = 20
    _BARRACKS = 21
    _MARINE = 48

    #Status
    _BUSY = 99
    #Lists for sorting the differnt types of units on the screen
    mines = []
    vespian = []
    commandCenters = []
    SCVs = []
    refineries = []
    barracks = []
    marines = []

    unitSelected = False

    def __init__ (self, obs, friendDims):
        self.selectedUnit = []

        self.mines.clear()
        self.vespian.clear()
        self.commandCenters.clear()
        self.SCVs.clear()
        self.refineries.clear()
        self.barracks.clear()
        self.marines.clear()

        minXFriend = friendDims[0]
        maxXFriend = friendDims[1]
        minYFriend = friendDims[2]
        maxYFriend = friendDims[3]

        if minXFriend == 0: #Command center is in the top-left
            self.buildIncrement = rand.random() * 9
        else:
            self.buildIncrement = rand.random() * -9

        # Get a list mapping of each unit 
        # Identify all mine and command center locations
        for mine in obs.observation.raw_units:
            if mine.x >= minXFriend and \
                mine.x < maxXFriend and \
                mine.y >= minYFriend and \
                mine.y < maxYFriend: #Is this unit part of the base start area? If so, add to the helper lists

                if mine.unit_type == self._MINERAL_FIELD:
                    self.mines.append(mine)
                elif mine.unit_type == self._VESPENE_GEYSER:
                    self.vespian.append(mine)
                elif mine.unit_type ==  self._REFINERY:
                    self.refineries.append(mine)
                elif mine.unit_type == self._COMMAND_CENTER:
                    self.commandCenters.append(mine)
                elif mine.unit_type == self._SCV:
                    self.SCVs.append(mine)
                elif mine.unit_type == self._BARRACKS and mine.health == 1000:
                    self.barracks.append(mine)    
                elif mine.unit_type == self._MARINE:
                    self.marines.append(mine)
    '''
    Abstract class to return the next macro step in the action
    '''

    def getNextStep (self, obs, macroAction, screenSector):
        pass

    '''
    Abstract method to determine if the action was successful and take follow on actions if it was not
    '''
    def errorCheck (self, obs, actionResult):
        pass

    '''
    Implemented Action to Select an SCV Use select Idle Unit instead
    '''
    #def selectSCV (self, obs):
    #    #Make a list of idle SCVs
    #    idleSCVs = [scv for scv in self.SCVs if scv.order_length == 0]
    #    randomSCVIdx = rand.randint(0, len(idleSCVs)-1)
    #    self.selectedUnit = idleSCVs[randomSCVIdx]
        #return actions.FunctionCall(self._SELECT_POINT, [[0], [self.selectedUnit.x, self.selectedUnit.y]]) #Select the SCV
        #return actions._RAW_FUNCTIONS.

    '''
     Implemented Action to Select an SCV. Use select Idle Unit instead
    '''
    #def selectMarine (self, obs):
    #    countdown = len(self.marines)
    #    while countdown > 0:
    #        #randomMarineIdx = rand.randint(0, len(self.marines)-1)
    #        self.selectedUnit = self.marines[countdown - 1]
    #        if self.selectedUnit.order_length != 0: #Marine is busy and not chewing crayons
    #            countdown -= 1
    #        else:
    #            return actions.FunctionCall(self._SELECT_POINT, [[0], [self.selectedUnit.x, self.selectedUnit.y]]) #Select the Marine
    #    else:
    #        return actions.FunctionCall(self._NO_OP, [])

    '''
    select Idle Unit
    '''
    def SelectIdleUnit (self, unitList, idle):
        if idle == True:
            idleUnits = [unit for unit in unitList if unit.order_length == 0] #Creates a list of all idle units
        else:
            idleUnits = unitList
        if len(idleUnits) == 0:
            return None
        idleIdx = rand.randint(0, len(unitList) - 1)
        if len(idleUnits) == 0:
            return actions.RAW_FUNCTIONS.no_op()
        else:
            self.selectedUnit = idleUnits[idleIdx]
            return idleUnits[idleIdx]
    
    
    '''
    Determine the number of macro steps required
    '''
    def numMacroSteps (self, actionIdx):
        if actionIdx == 0 or actionIdx == 1 or actionIdx == 2 or actionIdx == 3:
            return 2
        else:
            return 0

    def can_do (self, obs, action):
        return action[0] in obs.observation.available_actions