from pysc2.agents import base_agent
from pysc2.env import sc2_env, run_loop
from pysc2.lib import actions, features, units
from absl import app
import random as rand
import math
import time
import numpy as np
from MacroAction import MacroAction 
from BuildSupplyDepot import BuildSupplyDepot
from BuildMarineBarracks import BuildMarineBarracks
from BuildMarine import BuildMarine
from ScoutAction import ScoutEnemy
from ScreenSectors import ScreenSectors
from GameState import GameState
from MoveMarine import MoveMarine
from NeuralNetwork import NeuralNetwork
from ScenarioRecord import ScenarioRecord
from HistoryRecord import HistoryRecord
from RunMission import RunMission

class PyAgent(base_agent.BaseAgent):

    #Dictionary of Possible actions and whether they are avilable at this point in the game
    #For example, a supply depot cannot be built until 100 minerals are present. 
    #All are set to true and will be adjusted with each iteration of the game

    _POSSIBLE_ACTIONS = {"Build Supply Depot": True,
                         "Build Marine Barracks": True,
                         "Train Marine": True,
                         "Scout": True}

    nnTraining = True

    def step (self, obs, actionObj, macroAction, screenSector):
        super(PyAgent, self).step(obs) #Run step function from the Parent Class
        #Case statement to figure out what action we're implementing
        #if isinstance(actionObj, BuildSupplyDepot): #Build Supply Depot
        return actionObj.getNextStep(obs, macroAction, screenSector)
        #return actions.FunctionCall(actions.FUNCTIONS.no_op.id, [])

    def setTraining (self, switch):
        self.nnTraining = switch

    def getTrianing (self):
        return self.nnTraining


def main(unused_argv):
    agent = PyAgent()
    model = NeuralNetwork([10], 4)
    history = HistoryRecord() #Overall History of Training
    macroAction = MacroAction()
    _COMMAND_CENTER = 18
    _NUM_EPOCHS = 10
    _MARINE_POSSIBLE_MOVES = ["home",
                              "enemy",
                              "retreat-horizontal",
                              "retreat-vertical"]
    _ACTION_SPACING = 3
    _NUM_TRAINED_RUN = 1
    _NUM_BARRACKS = 2
    _NUM_SUPPLY_DEPOTS = 4
    _NN_RL_EPOCHS = 200

    try:
        while True: 
            with sc2_env.SC2Env(
                map_name="Simple64",
                players = [sc2_env.Agent(sc2_env.Race.terran),
                           sc2_env.Bot(sc2_env.Race.zerg,
                                       sc2_env.Difficulty.very_easy)],
                agent_interface_format = features.AgentInterfaceFormat(
                    feature_dimensions=features.Dimensions(screen=64, minimap=64),
                    action_space = actions.ActionSpace.RAW,
                    use_feature_units=False,
                    use_raw_units=True,
                    raw_resolution = 64),
                step_mul=8,
                game_steps_per_episode=0,
                visualize=True) as env:

                agent.setup(env.observation_spec(), env.action_spec())
                #timesteps = env.reset()
                #agent.reset()
                #agent.reward = 0

                #missionRunner = RunMission(timesteps, _NUM_BARRACKS, _NUM_SUPPLY_DEPOTS, env, agent)
                
                #Run the initial random training
                for epoch in range(0, _NUM_EPOCHS):
                    timesteps = env.reset()
                    agent.reset()
                    agent.reward = 0

                    missionRunner = RunMission(timesteps, _NUM_BARRACKS, _NUM_SUPPLY_DEPOTS, env, agent)
                    missionRunner.runMission(timesteps, model)
                    history.addHistoryRecord(missionRunner.getScenarioRecord())
                    missionRunner.clearScenarioRecord()
                    
                
                #Setup the NN for training
                history.convertToNPArray()
                model.initialize_model(len(history.modelInputs[0]))  
                #Train the NN
                model.trainModel(history)
                model.saveModel()

                for epoch in range(0, _NN_RL_EPOCHS):
                    #Run model inference and add the results to the history
                    timesteps = env.reset()
                    agent.reset()
                    agent.reward = 0
                    history.clearHistoryRecord() #Clear out the previous epochs. The model has already trained on them. 

                    missionRunner = RunMission(timesteps, _NUM_BARRACKS, _NUM_SUPPLY_DEPOTS, env, agent)
                    agent.setTraining(False)
                    missionRunner.runMission(timesteps, model)
                    history.addHistoryRecord(missionRunner.getScenarioRecord())
                    missionRunner.clearScenarioRecord()
                    
                    if len(history.history) > 0: #Sometimes the base gets destroyed before a marine is created :-()
                        #Setup the NN for training
                        history.convertToNPArray()
                        model.initialize_model(len(history.modelInputs[0]))  
                        # Train the NN
                        model.trainModel(history)
                        model.saveModel()
                   
                    
                    #while True:
                    #    #This experiment will only have the Neural Network learn how to fight. 
                    #    #I will script building the Barracks and Supply Depots. The AI will start issuing actions as soon as the first marine is built.

                    #    #Scripted Actions
                    #    gameState.updateState(timesteps[0])
                    #    #Adding this check to stop the SCVs from building after attack
                    #    if gameState.getNumBarracks() >= _NUM_BARRACKS and gameState.getNumSupplyDepots() >= _NUM_SUPPLY_DEPOTS:
                    #        _BASE_COMPLETE = True
                    #    minerals = gameState.getMinerals()
                    #    num_supply_depots_built = gameState.getNumSupplyDepots()
                    #    num_barracks_built = gameState.getNumBarracks()
                    
                    #    if gameState.canMoveMarine() and moveOrMarine == "move":
                    #        if armyMove.isNewOrder() == True:
                    #            model.generateCurrentState(timesteps[0])
                    #            if actionSpaceCounter == 0:
                    #                armyMove.generateAction(agent.getTrianing(), model, timesteps[0])
                    #                actionSpaceCounter = _ACTION_SPACING
                    #            else:
                    #                actionSpaceCounter -= 1
                    #            record.addStep(model.nnInputs, armyMove.getAction(), True)
                    #            armyMove.setMarineTags(timesteps[0])
                    #            #I want to move all marines before we do anything else
                                       
                    #        marineTag = armyMove.getMarineTag()
                    #        if marineTag == None:
                    #            actionObj = "No Action"
                    #        else:
                    #            actionObj = MoveMarine(timesteps[0], friendDims, gameState.getCCLocation(), _MARINE_POSSIBLE_MOVES[armyMove.getAction()], marineTag)
                    #        if len(armyMove.marineTags) == 0:
                    #            moveOrMarine = "marine" 
                    #    #Build _NUM_SUPPLY_DEPOTS First
                    #    elif num_supply_depots_built < _NUM_SUPPLY_DEPOTS and gameState.canBuildSupplyDepot() == True and _BASE_COMPLETE == False:
                    #        actionObj = BuildSupplyDepot(timesteps[0], friendDims)
                    #        num_supply_depots_built += 1
                    #    elif num_barracks_built < _NUM_BARRACKS and gameState.canBuildBarracks() == True and _BASE_COMPLETE == False:
                    #        actionObj = BuildMarineBarracks(timesteps[0], friendDims)
                    #        num_barracks_built += 1
                    
                    #    elif gameState.canBuildMarine() and moveOrMarine == "marine":
                    #        if len(timesteps[0].observation.production_queue) >= 3 or gameState.getNumMarines() > 10:
                    #            if rand.random() < 0.4:
                    #                moveOrMarine = "move"
                    #        actionObj = BuildMarine(timesteps[0], friendDims)
                    #    else:
                    #        actionObj = "No Action"

                    #    if actionObj != "No Action":
                    #        step_actions = [agent.step(timesteps[0], actionObj, macroAction, screenSectors)] #Need to complete the first step of the action
                    #    else:
                    #        step_actions = [actions.RAW_FUNCTIONS.no_op()]
                    #    timesteps = env.step(step_actions)

                    #    if timesteps[0].last():
                    #        record.normalizeReward(timesteps[0].reward) #Come up with the reward for all step actions
                    #        record.assignReward()
                    #        history.addHistoryRecord(record)
                    #        _BASE_COMPLETE = False
                    #        break

            

            #Run the predictions
            agent.setTraining(False)
                
            for run in range(_NUM_TRAINED_RUN):
                screenSectors = ScreenSectors(64)
                screenSectors.createSectors(4, 4)

                agent.setup(env.observation_spec(), env.action_spec())
                timesteps = env.reset()
                agent.reset()
                agent.reward = 0
                #Maintains a record for each epoch
                record = ScenarioRecord()

                #Find the command center
                for unit in timesteps[0].observation.raw_units:
                    if unit.unit_type == _COMMAND_CENTER:
                        command_center = unit
                        break

                #Determine which sector contains the Command Center and output the min, max grid coordinates
                friendDims = screenSectors.getFriendlyArea(command_center)

                #Constants for the initial game sequence
                _NUM_BARRACKS = 2
                _NUM_SUPPLY_DEPOTS = 4
                _BASE_COMPLETE = False
                gameState = GameState(timesteps[0])
                _AGENT_SLOW = 1 #Only take an action once every _AGENT_SLOW steps

                #Create a macro action object. This will be used to move all marines at once and nothing else
                armyMove = MacroAction()
                gameCounter = 0
                moveOrMarine = "marine"
                marineCount = 0
                #Initialize the first set of training epochs
                while True:
                    if gameCounter == _AGENT_SLOW:
                        gameCounter = 0
                        #This experiment will only have the Neural Network learn how to fight. 
                        #I will script building the Barracks and Supply Depots. The AI will start issuing actions as soon as the first marine is built.

                        #Scripted Actions
                        gameState.updateState(timesteps[0])
                        #Adding this check to stop the SCVs from building after attack
                        if gameState.getNumBarracks() >= _NUM_BARRACKS and gameState.getNumSupplyDepots() >= _NUM_SUPPLY_DEPOTS:
                            _BASE_COMPLETE = True
                        minerals = gameState.getMinerals()
                        num_supply_depots_built = gameState.getNumSupplyDepots()
                        num_barracks_built = gameState.getNumBarracks()
                    
                        if gameState.canMoveMarine() and moveOrMarine == "move":
                            if armyMove.isNewOrder() == True:
                                model.generateCurrentState(timesteps[0])
                                armyMove.generateAction(agent.getTrianing(), model, timesteps[0])
                                record.addStep(model.nnInputs, armyMove.getAction(), True)
                                armyMove.setMarineTags(timesteps[0])
                                #I want to move all marines before we do anything else
                                       
                            marineTag = armyMove.getMarineTag()
                            if marineTag == None:
                                actionObj = "No Action"
                            else:
                                actionObj = MoveMarine(timesteps[0], friendDims, gameState.getCCLocation(), _MARINE_POSSIBLE_MOVES[armyMove.getAction()], marineTag)
                            if len(armyMove.marineTags) == 0:
                                moveOrMarine = "marine" 
                        #Build _NUM_SUPPLY_DEPOTS First
                        elif num_supply_depots_built < _NUM_SUPPLY_DEPOTS and gameState.canBuildSupplyDepot() == True and _BASE_COMPLETE == False:
                            actionObj = BuildSupplyDepot(timesteps[0], friendDims)
                            num_supply_depots_built += 1
                        elif num_barracks_built < _NUM_BARRACKS and gameState.canBuildBarracks() == True and _BASE_COMPLETE == False:
                            actionObj = BuildMarineBarracks(timesteps[0], friendDims)
                            num_barracks_built += 1
                    
                        elif gameState.canBuildMarine() and moveOrMarine == "marine":
                            if len(timesteps[0].observation.production_queue) >= 3:
                                if rand.random() < 0.5:
                                    moveOrMarine = "move"
                            actionObj = BuildMarine(timesteps[0], friendDims)
                        else:
                            actionObj = "No Action"

                        if actionObj != "No Action":
                            step_actions = [agent.step(timesteps[0], actionObj, macroAction, screenSectors)] #Need to complete the first step of the action
                        else:
                            step_actions = [actions.RAW_FUNCTIONS.no_op()]
                        timesteps = env.step(step_actions)

                    else: #Counter has not expired
                        gameCounter += 1
                        step_actions = [actions.RAW_FUNCTIONS.no_op()]
                        timesteps = env.step(step_actions)

                    if timesteps[0].last():
                        break

    except KeyboardInterrupt:
        pass



if __name__ == "__main__":
    app.run(main)